/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#if !defined(_SOL_SOLARRAY)
#define _SOL_SOLARRAY

/**
  Associative dynamic array.
  */

class solarray {
 std::vector<dyn> data;
 static dyn nullelement;	
	
 public:

 int indexer;

 /**
   Return the number of elements
   */
 inline int size()   { return (int) data.size(); }

 /**
   Return true if the solsolarray holds no any element, false otherwise
   */
 inline bool empty() { return data.empty(); }

 void clear();
 void set(const solarray &);
 void set(_intArr);
 void set(_arr32);
 void set(_array64);
 void set(solarray *);
 void set(dyn);
 void set(char *[]);
 void set(std::vector<dyn> &);
 inline void set(std::vector<dyn> *a) { set(&a); }

 solarray();

 /**
   Build an associative solarray from an solarray of strings of chars.
   Indexes are generated from 0 to n.
   */
 solarray(char *[]);

 /**
   Build an solarray with a single element.
   */
 solarray(dyn);
 
 /**
  Clone
  */
 
 solarray(const solarray &);
 solarray(solarray *);
 solarray(std::initializer_list<dyn>);
 
 /** From a std array
  */

solarray(std::vector<dyn> &);
solarray(std::vector<dyn> *);


inline solarray(_intArr a) { set(a); }
inline solarray(_arr32 a) { set(a); }
inline solarray(_array64 a) { set(a); }

 /**
   Build a solarray from a typed indiced solarray.
   */
 
 ~solarray();


 void operator=(const solarray &);
 void operator=(solarray *);
 void operator=(dyn);
 void operator=(_intArr &);
 void operator=(_arr32 &);
 void operator=(_array64 &);
 void operator=(std::vector<dyn>);


 void add(solarray *);
 void add(const solarray &);

 //dyn get(int);

 inline solarray operator+(solarray x) { return operator+(&x); }
 solarray operator+(solarray *);
 solarray operator+(dyn);
 
 inline solarray operator-(solarray x) { return operator-(&x); }
 solarray operator-(solarray *);
 solarray operator-(dyn);

 
 inline solarray operator&(solarray x) { return operator&(&x); }
 solarray operator&(solarray *);
 
 inline solarray operator|(solarray x) { return operator|(&x); }
 solarray operator|(solarray *);
 
 inline solarray operator^(solarray x) { return operator^(&x); }
 solarray operator^(solarray *);

 inline solarray &operator+=(solarray x) { return operator+=(&x); }
 solarray &operator+=(solarray *);
 solarray &operator+=(dyn);
 
 void operator-=(solarray x) { operator-=(&x); }
 void operator-=(solarray *);
 void operator-=(dyn);
 
 inline bool operator==(solarray x) { return operator==(&x); }
 bool operator==(solarray *);
 bool operator==(bool);
 
 inline bool operator!=(solarray x) { return operator!=(&x); }
 bool operator!=(solarray *);
 bool operator!=(bool);
 
 inline bool operator<(solarray x) { return operator<(&x); }
 bool operator<(solarray *);
 
 inline bool operator>(solarray x) { return operator>(&x); }
 bool operator>(solarray *);
 
 inline bool operator>=(solarray a) { return !operator<(&a); }
 inline bool operator>=(solarray *a) { return !operator<(a); }
 
 inline bool operator<=(solarray b)  { return !operator>(&b); }
 inline bool operator<=(solarray *b) { return !operator>(b); }
 
 inline void operator&=(solarray x) { operator&=(&x); }
 void operator&=(solarray *);
 
 inline void operator|=(solarray x) { operator|=(&x); }
 void operator|=(solarray *);
 
 inline void operator^=(solarray x) { operator^=(&x); }
 void operator^=(solarray *);
 
 bool operator!();
 
 solarray &operator,(dyn);

 //===== search
 bool in(const dyn &);
 bool in(dyn *);
 
 signed int findValue(const dyn &);
 signed int findValue(dyn *);
 inline int find(dyn x)  { return findValue(x); }
 inline int find(dyn *x) { return findValue(x); }
 signed int findLastValue(const dyn &);
 signed int findLastValue(dyn *);
 inline int findLast(dyn x)  { return findLastValue(x); }
 inline int findLast(dyn *x) { return findLastValue(x); }
 
 //===== display
 
 void display();
 void display(int);

 //===== iterator

 /**
   Restart the iterator (at first element).
   */
 void reset();
 bool outOfBound();
 bool bound();
 bool inBound();
 /**
   Return the current element, and move index to next one.
   */
 dyn inc();

 /**
   Return the current element, and move index to previous one.
   */
 dyn dec();

 /**
   Return the currently indiced element.
   */
 dyn current();

 /**
   Return the key of the currently pointed out element.
   */
 text key();

 /**
   Return the key of the element at position in argument.
   */
 text key(int);

 /**
   Return the value of the currently pointed out element.
   */
 dyn value();


 /**
   Return the value of the element at position in argument.
   */
 dyn value(int);

 /**
   Set iterator at first element and return it.
   */
 dyn begin();

 /**
   Set iterator at last element and return it.
   */
 dyn end();

 /**
   Return the current indice of the iterator.
   */
 int index();

 //===== stack

 /**
   Push a dyn on top of the solarray, it becomes the last element
   and the size increases from one.
   */
 void push(const dyn &);
 void push(dyn *);
 void push(const char *);
 
 void pushUnic(dyn);

 /**
   Return the last element, and remove it from the solarray.
   If empty, return false.
   */
 dyn  pop();

 /**
   Return the first element and remove it.
   All remaining element are shift for one position. Indices are
   renumbered if integers.
   */
 dyn  shift();

 /**
   Insert an element at the first position.
   */
 void unshift(dyn);

 /**
   Insert an element at given position.
   The first parameter is the postion, the second one the element.
   */

 void insert(int, dyn);

 //===== ranges
 void erase(int, int);
 void erase(int);
 solarray slice(int, int, int);
 void splice(int, int, solarray, int);
 void splice(int, int, solarray *, int);
 void splice(int, int, int);
 void splice(int, int);
 void splice(int);
 void splice(int, int, dyn, int);
 void splice(int, int, bool, int);

 //===== transform

 /**
   Reverse the content of the solarray
   */
 solarray reverse();

 /**
   Sort the values of the solarray in ascending order.
   */
 void sort(int x = 0);
 void rSort(int x = 0);
 void swapData(int, int);
 void swapElem(int, int);
 void dynSort(int, int);
 void dynRSort(int, int); 

 /**
   Remove doubloons in values.
   */
 solarray unique();

 /**
   Return the values as an solarray. Keys are generated, from 0 to n.
   */
 solarray values();
 solarray keys();

 //===== operations
 void clone(solarray *);   // copy to another solarray
 void pack();  // for compatibility only

  /**
   Return the minimal value among elements.
   */
 dyn min();

 /**
   Return items selected by the function.
   */
 solarray map(bool (*)(dyn));

 /**
   Return the greatest values among elements.
   */
 dyn max();

 /**
   Calculate the sum of elements and return it.
   */
 double sum();

 /**
   Concatenate all the element (providing they are texts)
   and use the text given as argument as separator.
   */
 text join(const char *);

 int nomatch();

 //===== I/O

 /**
   Load a file into the solarray. The argument is the filename.
   Each line of the file become an element. The end of line code
   (10, 13 or the twice) are stored also.
   */
 void load(const char *);

 /**
   Save the whole solarray into the file, the name of which is given as
 argument. An end of line must be added to each element if necessary.
  */
 bool store(const char *);


 solarray *to(int);

 dyn &operator[](int);

 dyn &at(int);
 dyn &at(dyn);
 
 void setAt(int, dyn *);
 void setAt(int, dyn);

 char *toString(char *s=", ");
 text toText(char *s=", ");
 void boundError(const text &);

};


inline solarray operator+(solarray *a, const solarray &b) { return a->operator+(b); }


std::ostream& operator<< (std::ostream& out, solarray *);
std::ostream& operator<< (std::ostream& out, solarray);

#endif

