/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>

#include <algorithm>
#include <iterator>
#include <vector>
#include <iostream>
#include <string>

using namespace std;

#ifdef UNIX
#include <unistd.h>
#else
#define strcasecmp strcmpi
#include <io.h>
#endif
#include <math.h>

#include "text.hpp"
#include "textarr.hpp"
#include "intarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"
#include "dom.hpp"
#include "lximport.h"
#include "sax.hpp"


extern char *substr(const char *, int, int);

#define TABTXT_LEN 48
#define _DOM_RESET_INDEX 1

const char *dom::tabtxt = "                                                 ";

dom *lxdom = new dom();

const int dom::XML = 0;
const int dom::XSL = 1;
const int dom::SVG = 2;
const int dom::XUL = 3;
const int dom::LIGHT = 4;
const int dom::TREE = 5;

const int dom::LIBXML = 0;
const int dom::EXPAT = 1;

//============== XCOMMENT

xcomment::xcomment()
{
  data = "";
}


xcomment::xcomment(xcomment *x)
{
  data = x->data;
}


void xcomment::set(xcomment *x)
{
  data = x->data;
}


//============== XNODE

xnode::xnode()
{
  type = 0;
  parent = 0;
  child = 0;
  next = 0;
  lastChild = 0;
  owner = NULL;
}

xnode::xnode(xnode *x)
{
  set(x);
}

void xnode::set(xelement *x) { element = x; }
void xnode::set(xscript *x)  { statement = x; }
void xnode::set(xcomment *x) { comment = x; }
void xnode::set(xnode *x)
{
  type = x->type;
  switch(type)
  {
   case ELEMENT:
        element = new xelement(x->element);
        break;
   case SCRIPT:
        statement = new xscript(x->statement);
        break;
   case COMMENT:
        comment = new xcomment(x->comment);
        break;
   default:
        break;
  }

  parent = 0;
  child = 0;
  next = 0;
  lastChild = 0;
  owner = NULL;
 
  return;
}

/*
char *xnode::getName()
{
	char *name;
	
	switch(x->type)
	{
	case ELEMENT:
        name = element->name;
        break;
	case SCRIPT:
        name = statement->name;
        break;
	case COMMENT:
        name = commen->name;
        break;
	default:
		name = "";
        break;
	}
	return name;
}
*/


//============== X ATTRIBUTE

xattribute::xattribute(const text &a, const text &b)
 {
    name = a;
    value = b;
 }

void xattribute::set(const text &a, const text &b)
 {
    name = a;
    value = b;
 }

//============== X ELEMENT

xelement::xelement()
{
  name = "";
  data = "";
  index = 0;
}


xelement::xelement(const char *n)
{
  setName(n);
  data = "";
  index = 0;
}

xelement::xelement(xelement *x)
 { 
 	set(x); 
 }

void xelement::setName(const char *s)
{
	int l = strlen((char *) s);
	name = (char *) new char[l+1]; 
	strcpy(name, (char *) s);
}

void xelement::setData(const char *s)
{
	int l = strlen(s);
	data = (char *) new char[l+1];
	strcpy(data, s);
}



void xelement::set(xelement *x)
{
  int i, l;
  text n, v;
  setName(x->name);
  setData(x->data);

  l = x->attributes.size();
  for(i = 0; i < l; i++)
  {
    n = x->attributes[i]->name;
    v = x->attributes[i]->value;
    addAttribute(n, v);
  }
}


void xelement::assign(xelement *x)
{
	int i, l;
	text n, v;

	setData(x->data);

	l = x->attributes.size();
	for(i = 0; i < l; i++)
	{
		n = x->attributes[i]->name;
		v = x->attributes[i]->value;
		addAttribute(n, v);
	}
}


void xelement::addAttribute(xattribute *xa)
 {
    attributes.push_back(xa);
 }

void xelement::addAttribute(const text &a, const text &b)
 {
   xattribute *xa = new xattribute(a, b);
   attributes.push_back(xa);
 }

xattribute *xelement::getAttribute() { return attributes[index]; }

bool xelement::hasAttribute(const char *name)
{
	int len = (int) attributes.size();
	for(int i = 0; i < len; i++)
	{
		if(strcasecmp(name, attributes[i]->name.data) == 0)     return true;
	}
	return false;
}


char *xelement::getValue(const char *name)
{
  int len = (int) attributes.size();
  for(int i = 0; i < len; i++)
  {
     if(strcasecmp(name, attributes[i]->name.data) == 0)
     {
        return attributes[i]->value.data;
     }
  }
  return "";
}

void xelement::setValue(const char *name, const char *value)
{
  int len = (int) attributes.size();
  for(int i = 0; i < len; i++)
  {
     if(strcasecmp(name, attributes[i]->name.data) == 0)
     {
        attributes[i]->value = value;
     }
  }
}


char *xelement::nextName()
{
  if(index < (int) attributes.size())
  {
    xattribute *xa = attributes[index];
    return xa->name.data;
  }
  return "";
}

char *xelement::nextValue()
{
  if(index < (int) attributes.size())
  {
    xattribute *xa = attributes[index];
    return xa->value.data;
  }
  return "";
}

xelement *xelement::begin() { index = 0; isFound = true; return this; }

xelement *xelement::inc()
{
  if((index + 1) >= (int) attributes.size())
  {
    isFound = false;
    return this;
  }
  isFound = true;
  index++;
  return this;
}

bool xelement::found()
{
  return isFound == true;
}


void xelement::display()
{
   xattribute *att;

   printf("%s", name);

   for(unsigned int i = 0; i < attributes.size(); i++)
   {
      att = attributes[i];
      printf(" %s=%s", att->name.data, att->value.data);
   }

   if(strlen(data) > 0)
   {
     puts("");
     printf("    \"%s\"\n", data);
   }
   else
   {
     puts("");
   }
  printf("/%s\n", name);

}

//============== X SCRIPT

xscript::xscript()
{
  index = 0;
  tab = 0;
}


xscript::xscript(xscript *x)
{
  index = 0;
  tab = 0;

  int n = x->lines.size();
  for(int i = 0; i < n; i++)
  {
    lines.push_back(x->lines[i]);
  }
}


void xscript::set(xscript *x)
{
  index = 0;
  tab = 0;

  int n = x->lines.size();
  for(int i = 0; i < n; i++)
  {
    lines.push_back(x->lines[i]);
  }
}


void xscript::push(text *t)
{
  lines.push_back(t);
  tabs.push_back(tab);
}


void xscript::push(const string &s)
{
  lines.push_back(new text(s.c_str()));
  tabs.push_back(tab);
}


void xscript::reset()
{
  index = 0;
}


int xscript::getTab() { return tabs[index]; }


const char *xscript::next()
{
  if(index >= size()) return "";
  return lines[index++]->data;
}

bool xscript::again()
{
  return(index < size());
}

void xscript::load(const char *fname)
{
  FILE *fp;
  char *tst;
  char buf[32000];
  fp = fopen(fname, "r");
  if(fp == NULL) return;
  while(1)
  {
    tst = fgets(buf, 65536, fp);
    if(tst == NULL) break;
    push(new text(buf));
  }
  fclose(fp);
}

//============= DOM


dom::dom()
{
   clear();
}


void dom::assign(dom *original)
{
  xnode *sno;
  int srcindex;

  if(size() > 1)
  {
    clear();
  }
  //index = _DOM_RESET_INDEX;
  index = 0;

  //cout << "set nodes:" << nodes.size() << endl;

  sno = original->nodes[0];
  srcindex = sno->child;
  if(srcindex > 0)
  {
    sno = original->nodes[srcindex];
    pushChild(sno);
  }
}


dom::dom(const char *s)
{
  clear();
  setName(s);
}

dom::~dom() { }


void dom::addHeader(const char *str)
{
  header.push_back(new text(str));
}

void dom::clear()
{
  nodes.clear();
  header.clear();

  top = 0;
  index = 0;
  isFound = true;

  xnode *x = new xnode();

  xelement *xe = new xelement();
  xe->name = "document";
  x->set(xe);
  x->owner = this;
  nodes.push_back(x);

  element = xe;   // prevent crash
  node = x;
  top++;
}


// no previous node, is a first child

bool dom::isElder(int idx)
{
	//printf("current %d\n", idx);
	node = nodes[idx];         				// get the current node if idx is index
	//printf("parent %d\n", node->parent);
	node = nodes[node->parent];		// get parent index
	//printf("child%d ?%d\n", node->child, node->child == idx );
	return node->child == idx;			// get first child of parent and compare
}	


//=== BUILDING

int dom::addNext(xnode *x, int type)
 {
   int parent;

   node = nodes[index];        // gets the index node, becomes previous
   node->next = top;           // sets this node to be the next for it
   parent = node->parent;

   node = nodes[parent];       // gets the parent
   node->lastChild = top;      // gives it the last child, index

   x->type = type;             // sets the type of content into the node
   x->parent = parent;         // the parent of this node is that of previous one
   x->owner = this;
   nodes.push_back(x);         // pushes this node on the vector

   index = top;                // this node becomes the index number
   top++;                      // changes the top for another node
   return index;               // returns the index of the new node
 }


// if the parent has already a child
// search for the last child
// add new node next this last node
// and change lastChild to the last node
// require:
// index points out the parent node

int dom::addChild(xnode *x, int type)
 {

   x->parent = index;            // gives last node a parent
   x->type = type;              // and the type
   x->owner = this;
   nodes.push_back(x);

   node = nodes[index];          // gets the parent
   if(node->child)
   {
     x->child = node->child;
   }
   node->child = top;           // this entry becomes new child
   index = top;                // prepare next entry, index on current
   top++;                     // points out future entry
   return index;
 }


// get a branch from this tree or another tree
// put it on top of stack

// copy the xnode, perhaps from another tree
// - copy the content
// - set this tree the owner
// - links are set by addNext

void dom::copyNext(xnode *x)
 {
   addNext(new xnode(x), x->type);
 }


void dom::copyChild(xnode *x)
 {
   addChild(new xnode(x), x->type);
 }




void dom::pushNext(xnode *xo)
{
  int srchild, srcnext;
  dom *source = xo->owner;
  int savindex;

loop:
  //copyNext(xo);
  addNext(new xnode(xo), xo->type);
  srchild = xo->child;
  srcnext = xo->next;
  savindex = index;

  if(srchild) {
    pushChild(source->nodes[srchild]);
  }

  if(srcnext) {
    index = savindex;
    xo = source->nodes[srcnext];
    goto loop;
  }
}

/**
      Push as child of the node pointed out by index
  */

void dom::pushChild(xnode *xo)
{
  int srchild, srcnext;
  int savindex;
  dom *source = xo->owner;

  //copyChild(xo);    // copy as child of index
  addChild(new xnode(xo), xo->type);

loop:
  srchild = xo->child;
  srcnext = xo->next;
  savindex = index;

  if(srchild) {
    xo = source->nodes[srchild];
    pushChild(xo);
  }

  if(srcnext) {
    xo = source->nodes[srcnext];
    //pushNext(xo);
    //copyNext(xo);
    index = savindex;
    addNext(new xnode(xo), xo->type);
    goto loop;
  }
}


// put a branch at top of stack
// and link as last node inside top node

void dom::push(xnode *xo)
{

  //dom *source = xo->owner;
  xnode *xn = nodes[0];
  index = xn->child;
  xn = nodes[index];
  index = xn->lastChild;

  pushNext(xo);

}


// set a node the successor of a pointed out one
// - no previous successor, call add next
// - a successor:
// * the successor must become the successor of the new one
// * previous successor remains last child

void dom::setNext(xnode *xn)
{
  int next;
  int ctop = top;

  //xelement *xe = xn->element;
  //std::cout << "setnext " << xe->name << " child=" << xn->child << std::endl;

  node = nodes[index];
  next = node->next;

  pushNext(xn);  // push recursively by reading xnodes
  
  if(next == 0) return;  // if current has no successor, add just

  // if current had a successor,
  // put it into next fields of node added at (old) top of stack

  node = nodes[ctop];
  node->next = next;
  index = ctop;
}



void dom::setNext(xelement *xe)
{
   xnode *xn = new xnode();
   xn->set(xe);
   setNext(xn);
}

/*
  - no previous, set child (lastChild set by called function)
  - if previous child, adds as child and attaches previous as next
    * lastChild unchanged
*/

void dom::setChild(xnode *xn)
{
  int child;
  int ctop = top;

  node = nodes[index];      // get pointed out node
  child = node->child;      // has it a child?
  node->child = 0;          // reset it anyway to prevent this added as next

  pushChild(xn);            // push recursively by reading xnodes
  
  if(child == 0) return;    // if current has no child, that's all

  // if current had a child,

  node = nodes[ctop];       // the current node is at the top (previous)
  node->next = child;       // previous child becomes next to this one
  index = ctop;

}

void dom::setChild(xelement *xe)
{
   xnode *xn = new xnode();
   xn->set(xe);
   setChild(xn);
}






/**
   When tag closed, or at user call
   points out the successor of the parent
   - get the parent
   - get its successor
   - if no successor return
   - otherwise get its parent
   - and get its last child
  */

dom *dom::up()
{
  isFound = false;
  if(index)
  {
    node = nodes[index];      // gets last entered node
    index = node->parent;

    node = nodes[index];      // move up to its parent
    if(node->next == 0)       // if parent has no follower, link to it
    {
      return this;
    }

    index = node->parent;     // otherwise gets parent of the parent
    node = nodes[index];      //
    index = node->lastChild;  // points out its last child to link it
    if(index) isFound = true;
  }
  return this;
}


dom *dom::down()
{
   isFound = false;
   if(index > -1)
   {
     node = nodes[index];
     index = node->child;
     if(index > 0) isFound = true;
   }
   return this;
}

dom *dom::setIndex(int i)
{
   index = i;
   return this;
}


int dom::addAttribute(xattribute *xa)
 {
   node = nodes[index];
   if(node->type == ELEMENT)
   {
     element = node->element;
     element->addAttribute(xa);
   }
   return top - 1;
 }

int dom::addAttribute(int node, xattribute *xa)
 {
   index = node;
   return addAttribute(xa);
 }

int dom::addAttribute(const char *a, const char *b)
 {
   xattribute *xa = new xattribute(text(a), text(b));
   return addAttribute(xa);
 }


//=== ELEMENT

int dom::addNext(xelement *xe)
 {
   xnode *x = new xnode();     // creates a node
   x->set(xe);                 // gives it a content, here this is an element
   x->owner = this;            // gives address of the dom
   return addNext(x, ELEMENT); // adds the node to vector
 }

int dom::addNext(int n, xelement *xe)
 {
   index = n;
   return addNext(xe);
 }

int dom::addChild(xelement *xe)
 {
   xnode *x = new xnode();
   x->set(xe);
   x->owner = this;
   return addChild(x, ELEMENT);
 }

int dom::addChild(int n, xelement *xe)
 {
   index = n;
   return addChild(xe);
 }

char *dom::getElementName()
{
   node = nodes[index];
   if(node->type == ELEMENT)
   {
     return node->element->getName();
   }
   return "";
}


//=== SCRIPT

int dom::addNext(xscript *xs)
{
   xnode *x = new xnode();
   x->set(xs);
   x->owner = this;
   return addNext(x, SCRIPT);
}

int dom::addChild(xscript *xs)
 {
   xnode *x = new xnode();
   x->set(xs);
   x->owner = this;
   return addChild(x, SCRIPT);
 }


// load the script file and fill the script with the content

void dom::addScriptChild(const char *fname)
{
	xscript *x = new xscript();
	x->load(fname);
	addChild(x);
}

void dom::addScriptNext(const char *fname)
{
	xscript *x = new xscript();
	x->load(fname);
	addNext(x);          // add a node with the script inside
}


//=== COMMENT

void dom::addComment(const char *str)
{
  xcomment *xc = new xcomment();
  xc->set(str);
  addNext(xc);
}

int dom::addNext(xcomment *xc)
{
   xnode *x = new xnode();
   x->set(xc);
   x->owner = this;
   return addNext(x, COMMENT);
}

int dom::addChild(xcomment *xc)
 {
   xnode *x = new xnode();
   x->set(xc);
   x->owner = this;
   return addChild(x, COMMENT);
 }


// adds a line to a script

void dom::addLine(const string &s)
{
  statement->push(s);        // should be a pointer onto a script
}


//============ OUTPUT

// disp or save a line

void dom::send(const char *txt)
{
  if(fileFlag)
  {
    fputs(txt, domfile);
  }
  else
  {
    printf("%s", txt);
  }
}

void dom::send(char c)
{
  if(fileFlag)
  {
    fputc(c, domfile);
  }
  else
  {
    printf("%c", c);
  }
}

void dom::sendText(const char *txt)
{
  if(strlen(txt) > 0)
  {
    if((txt[0] == '\"') || (txt[0] == '\''))
    {
      send(txt);
      return;
    }
  }
  send("\"");
  send(txt);
  send("\"");
}

void dom::indent()
 {
    int i = tab * 2;

    if(i > TABTXT_LEN) i = TABTXT_LEN;
    send(substr(tabtxt, 0, i));
 }

// disp the starting of a node

void dom::openTag(xelement *xe, int child)
 {
   xattribute *att;

   indent();
   if(mode == XML) send('<');
   send(xe->name);

   for(unsigned int i = 0; i < xe->attributes.size(); i++)
   {
      att = xe->attributes[i];
      send(" ");
      send(att->name.data);
      send("=");
      sendText(att->value.data);
   }

   if(strlen(xe->data) > 0)
   {
     if(mode == XML) send('>');
     send('\n');
     //indent();
     if(mode == LIGHT) send('\"');
     if(xe->data != NULL)
     {
       send(xe->data);
     }
     if(mode == LIGHT) send('\"');
     send('\n');
   }
   else
   {
     if(child != 0 || htmlFlag)
     {
       if(mode == XML) send('>');
       send('\n');
     }
   }

 }

// disp the ending of a node

void dom::closeTag(xelement *xe, int child)
 {
   if(child || (strlen(xe->data) > 0) || htmlFlag)
   {
    indent();
    if(mode == XML) send('<');
    send('/');
    send(xe->name);
    if(mode == XML) send('>');
    send('\n');
   }
   else
   {
    if(mode == XML) send(" />\n");
    else            send(" /\n");
   }
 }


void dom::outData(const char *name, int num)
{
  if(num)
   {
     send(name);
     send(": ");
     node = nodes[num];
     if(node->type == ELEMENT)
     {
      element = node->element;
      send(element->name);
     }
     send("\n");
   }
}


void dom::dump()
 {
   int parent;
   int child;
   int next;
   int last;
   char buffer[16];

   fileFlag = false;
   tab = 1;
   for(unsigned int i = 0; i < (unsigned) top; i++)
   {
     node = nodes[i];
     switch(node->type)
     {
       case ELEMENT:
            element = node->element;
            parent = node->parent;
            child = node->child;
            next = node->next;
            last = node->lastChild;
            send("ELEMENT:");
            send(element->name);
            send("\n");
            outData("parent", parent);
            outData("child ", child);
            outData("next  ", next);
            outData("last  ", last);
            send("\n");
            break;
       case SCRIPT:
            send("SCRIPT ");
            statement = node->statement;
            sprintf(buffer, "%d", statement->size());
            send(buffer);
            send(" lines\n");
            break;
       case COMMENT:
            send("COMMENT length=");
            comment = node->comment;
            sprintf(buffer, "%d", comment->data.length());
            send(buffer);
            send("\n");
            break;
       default:
            break;
     }
   }
 }


void dom::subout(int no, int t)
 {
   int child;

   union {
    xelement *xe;
    xscript *xs;
    xcomment *xc;
   };

 follow:
   node = nodes[no];        // gets the index node
   child = node->child;     // gets its child
   no = node->next;         // gets the follower
   switch(node->type)
   {
    case ELEMENT:
     xe = node->element;
     tab = t;
     openTag(xe, child);
     if(child)
     {
       subout(child, tab + 1);
     }
     tab = t;
     closeTag(xe, child);
     break;
   case SCRIPT:
     xs = node->statement;
     if(xs != NULL)
     {
       xs->reset();
       while(xs->again())
       {
         tab = xs->getTab();
         indent();
         send(xs->next());
       }
     }
     break;
   default:
     xc = node->comment;
     indent();
     if(mode == XML) send("<");
     send("comment data=");
     send(xc->data.data);
     /*
     if(xc->sign.length() > 0)
     {
      indent();
      send(" code=");
      sendText(xc->sign.data);
     }
     */
     if(mode == XML) send(" />\n");
     else            send(" /\n");
     break;
  }

   if(no) goto follow;              // loops
 }

void dom::outHeader()
{
  for(int i = 0; i < (int) header.size(); i++)
  {
    send(header[i]->data);
    send("\n");
  }
}

void dom::save(const char *fname, int m)
{
  if(empty()) return;

  fileFlag = true;
  this->mode = m;
  domfile = fopen(fname, "wb");
  outHeader();
  subout(1,0);
  fclose(domfile);
}

void dom::append(const char *fname, int m)
{
  if(empty()) return;

  fileFlag = true;
  this->mode = m;
  domfile = fopen(fname, "a");
  outHeader();
  subout(1,0);
  fclose(domfile);
}


void dom::display(int m)
 {
  if(empty())
  {
    std::cout << "Empty\n";
    return;
  }

  fileFlag = false;
  this->mode = m;
  outHeader();
  subout(1, 0);
 }

// return the content if name = "data" or search for an attribute

char *dom::getValue(const char *name)
{
  node = nodes[index];
  if(node->type != ELEMENT) return "";
  element = node->element;

  return element->getValue(name);
}

// change value at pointer

void dom::setValue(const char *name, const char *value)
{
  node = nodes[index];
  if(node->type != ELEMENT) return;
  element = node->element;
  element->setValue(name, value);
}


// Get the content of currently pointed out node

char *dom::getData()
{
   node = nodes[index];
   if(node->type != ELEMENT) return "";
   element = node->element;
   //printf("node index=%d, tag:%s data(%s)\n", index, element->name.data, element->data.data);

   return element->data;
}


// Set the content of currently pointed out node

void dom::setData(const char *str)
{
   node = nodes[index];
   if(node->type == ELEMENT)
   {
     element = node->element;
     element->setData(str);
   }
}




void dom::unknown(const char *name)
{
  std::cout << name << " element not found\n";
  return;
}

void dom::noElement(const char *name)
{
  std::cout << "No more element, for " << name << "\n";
  return;
}


/**
   The at method search for a sub tag
   If the sub-tag doesn't match the name,
   search for following sub-tab.
   - get the child
   - if no child exit
   - compare the name of the child element
   - if differs, gets the next one
   - if no next exit
  */

dom *dom::at(const char *name)
{
  int idx;
  isFound = false;

  //std::cout << "<SEARCH: " << name << " index=" << index << ">" << std::endl;

  node = nodes[index];
  index = node->child;      // get the child
  //std::cout << "child= " << idx << std::endl;
  if(index == 0)            // not child? exits
  {
    noElement(name);
    return this;
  }

  idx = index;

loop:
  node = nodes[idx];              // gets the child/next of child node
  if(node->type == ELEMENT)
  {
    element = node->element;
    if(strcmp(element->name, name) == 0)
    {
      index = idx;
      isFound = true;
      return this;
    }
  }
  idx = node->next;
  if(idx) goto loop;

  unknown(name);
  return this;
}

// return the content of an element having as arguments
// the attribute and its value

bool dom::compare(const char *name, const char *value)
{
  int len;
  xattribute *xa;

  node = nodes[index];
  if(node->type == ELEMENT)
  {
    element = node->element;
    len = (int) element->attributes.size();

    for(int i = 0; i < len; i++)
    {
      xa = element->attributes[i];
      if(xa->name.compare(name) == 0)
      {
        if(xa->value.compare(value) == 0) return true;
      }
    }
  }

  return false;
}


dom *dom::at(const char *name, const char *value)
{
  isFound = true;
  int idx;

loop:
  if(compare(name, value)) return this;
  node = nodes[index];
  idx = node->next;
  if(idx)
  {
    index = idx;
    goto loop;
  }

  isFound = false;
  return this;
}



//========================== READING

int dom::getChild()
{
  node = nodes[index];
  return node->child;
}

int dom::getNext()
{
  node = nodes[index];
  return node->next;
}

bool dom::isElement()
{
  node = nodes[index];
  return (node->type == ELEMENT);
}

bool dom::isScript()
{
  node = nodes[index];
  return (node->type == SCRIPT);
}

bool dom::isComment()
{
  node = nodes[index];
  return (node->type == COMMENT);
}


//========================== ITERATOR

dom *dom::start()
{
  index = 0;
  return this;
}

dom *dom::reset()
{
  index = _DOM_RESET_INDEX;
  isFound = true;
  return this;
}

dom *dom::begin()
{
  int parent;
  node = nodes[index];
  parent = node->parent;
  index = nodes[parent]->child;
  return this;
}


dom *dom::end()
{
  int parent;

  node = nodes[index];
  parent = node->parent;
  index = nodes[parent]->lastChild;
  return this;
}


dom *dom::inc()
{
  if(index == 0) { isFound = false; return this; }

  node = nodes[index];
  if(node->type == ELEMENT)
  {
    element = node->element;
  }

  index = node->next;
  if(index == 0)
  {
     isFound = false;
  }
  else
  {
     isFound = true;
  }

  return this;
}


dom *dom::dec()
{
  if(index == 0) { isFound = false; return this; }

  int orig = index;

  index = node->parent;
  node = nodes[index];
  index = node->child;

loop:
  node = nodes[index];
  if(orig == index)
  {
    isFound = false;
    return this;
  }

  index = node->next;
  if(index == orig)
  {
     node = nodes[index];
     if(node->type == ELEMENT)
     {
       isFound = true;
       return this;
     }
  }
  goto loop;

}



xnode *dom::getNode()
{
  xnode *xn = new xnode(nodes[index]);
  //xelement *xe = xn->element;
  //std::cout << "getnode " << xe->name << " child=" << xn->child << std::endl;
  return xn;
}

xelement *dom::getElement()
{
  node = nodes[index];
  if(node->type == ELEMENT)
  {
    return node->element;
  }

  return new xelement();
}

xscript *dom::getScript()
{
  node = nodes[index];
  if(node->type == SCRIPT)
  {
    return node->statement;
  }

  return new xscript();
}

xcomment *dom::getComment()
{
  node = nodes[index];
  if(node->type == COMMENT)
  {
    return node->comment;
  }

  return new xcomment();
}




void dom::remove()
{
  int next;
  int indice;
  int parent;

  node = nodes[index];
  next = node->next;

  parent = node->parent;
  node = nodes[parent];
  indice = node->child;             // gets child of parent
  if(indice == index)               // this is the child
  {
    node->child = next;            // successor becomes child of parent
    return;
  }

  // otherwise gets the child
loop:
  node = nodes[indice];             // gets the node, child first, next then
  indice = node->next;              // gets the successor
  if(indice == 0)                   // no successor
  {
    return;
  }
  if(indice == index)               // the successor it the pointed out node
  {
    node->next = next;              // got at start
    return;
  }
  goto loop;
}


//===============


bool dom::load(bool (*fun) (const char *fname), const char *fname)
{
  lxdom = this;
  return fun(fname);
}

bool dom::load(const char *fname)
{
	lxdom = this;
	return SaxImport(fname);
}


#define _DEBUG
#ifdef DEBUG

int main(int argc, char **argv)
{

   dom *dt = new dom();

   sax::load(dt->build, "xuldemo.xml");
   dt->display();

}

#endif

