/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <ctype.h>
#include <sys/stat.h>
#include <time.h>
#include <dirent.h>

#include <iostream>
#include <vector>

#ifndef VISUALC
#include <unistd.h>
#else
#include <io.h>
#include <errno.h>
#include <direct.h>
#endif


#include "text.hpp"
#include "textarr.hpp"
#include "intarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"
#include "fun.hpp"

// Start of replacement of libstd

#if defined(MSVC) | defined(VISUALC)

#define WIN32_LEAN_AND_MEAN
#define SUFFIX  "*"
#define	SLASH	"\\"

/*
  Returns a pointer to a DIR structure filled in at directory beginning.
 */

DIR *opendir (const char *szPath)
{
  DIR *nd;
	
  errno = 0;

  if (!szPath)
    {
      errno = EFAULT;
      return (DIR *) 0;
    }

  if (szPath[0] == '\0')
    {
      errno = ENOTDIR;
      return (DIR *) 0;
    }

  /* Attempt to determine if the given path really is a directory. */

/*
  struct ffblk data;
  int res = findfirst(szPath, &data, _A_SUBDIR);
  if( (res == -1) || !(data.ff_attrib & FA_DIREC)  )
  {
      errno = ENOTDIR;
      return (DIR *) 0;
  }
*/
  struct _finddata_t data;
  data.attrib = 0;

  _findfirst(szPath, &data);

  if(data.attrib & _A_SUBDIR);
  else
  {
      errno = ENOTDIR;
      return (DIR *) 0;
  }


  /* Allocate space for DIR structure plus the directory path */

  nd = (DIR *) GC_malloc_atomic(sizeof (DIR) + strlen (szPath) + strlen (SLASH) + strlen (SUFFIX));
  //nd = (DIR *) new char[sizeof (DIR) + strlen (szPath) + strlen (SLASH) + strlen (SUFFIX)];      
  
  if(!nd)
  {
      /* Error, out of memory. */
      errno = ENOMEM;
      return (DIR *) 0;
  }

  /* Create the search expression. */
  strcpy (nd->dd_name, szPath);

  /* Add on a slash if the path does not end with one. */
  if (nd->dd_name[0] != '\0' &&
      nd->dd_name[strlen (nd->dd_name) - 1] != '/' &&
      nd->dd_name[strlen (nd->dd_name) - 1] != '\\')
    {
      strcat (nd->dd_name, SLASH);
    }

  /* Add on the search pattern */
  strcat (nd->dd_name, SUFFIX);

  /* Initialize handle to -1 so that a premature closedir doesn't try
   * to call _findclose on it. */
  nd->dd_handle = -1;

  /* Initialize the status. */
  nd->dd_stat = 0;

  /* Initialize the dirent structure. ino and reclen are invalid under
   * Win32, and name simply points at the appropriate part of the
   * findfirst_t structure. */
  nd->dd_dir.d_ino = 0;
  nd->dd_dir.d_reclen = 0;
  nd->dd_dir.d_namlen = 0;
  nd->dd_dir.d_name = nd->dd_dta.name;

  return nd;
}


/*
   Return a pointer to a dirent structure filled in with next entry.
 */

struct dirent *readdir (DIR * dirp)
{
  errno = 0;

  /* Check for valid DIR struct. */
  if (!dirp)
    {
      errno = EFAULT;
      return (struct dirent *) 0;
    }

  if (dirp->dd_dir.d_name != dirp->dd_dta.name)
    {
      /* The structure does not seem to be set up correctly. */
      errno = EINVAL;
      return (struct dirent *) 0;
    }

  if (dirp->dd_stat < 0)
    {
      /* We have already returned all files in the directory
       * (or the structure has an invalid dd_stat). */
      return (struct dirent *) 0;
    }
  else if (dirp->dd_stat == 0)
    {
      /* We haven't started the search yet. */
      /* Start the search */
      dirp->dd_handle = _findfirst (dirp->dd_name, &(dirp->dd_dta));

  	  if (dirp->dd_handle == -1)
	{
	  /* Whoops! Seems there are no files in that
	   * directory. */
	  dirp->dd_stat = -1;
	}
      else
	{
	  dirp->dd_stat = 1;
	}
    }
  else
    {
      /* Get the next search entry. */
      if (_findnext (dirp->dd_handle, &(dirp->dd_dta)))
	{
	  /* We are off the end or otherwise error. */
	  _findclose (dirp->dd_handle);
	  dirp->dd_handle = -1;
	  dirp->dd_stat = -1;
	}
      else
	{
	  /* Update the status to indicate the correct
	   * number. */
	  dirp->dd_stat++;
	}
    }

  if (dirp->dd_stat > 0)
    {
      /* Successfully got an entry. Everything about the file is
       * already appropriately filled in except the length of the
       * file name. */
      dirp->dd_dir.d_namlen = strlen (dirp->dd_dir.d_name);
      return &dirp->dd_dir;
    }

  return (struct dirent *) 0;
}


/*
   closedir
   Frees up resources allocated by opendir.
 */
int closedir (DIR * dirp)
{
  int rc;


  errno = 0;
  rc = 0;


  if (!dirp)
    {
      errno = EFAULT;
      return -1;
    }


  if (dirp->dd_handle != -1)
    {
      rc = _findclose (dirp->dd_handle);
    }

  // useless with GC, causes errors
  //free (dirp);

  return rc;
}

/*
 * rewinddir
 *
 * Return to the beginning of the directory "stream". We simply call findclose
 * and then reset things like an opendir.
 */
void rewinddir (DIR * dirp)
{
  errno = 0;

  if (!dirp)
    {
      errno = EFAULT;
      return;
    }

  if (dirp->dd_handle != -1)
    {
      _findclose (dirp->dd_handle);
    }

  dirp->dd_handle = -1;
  dirp->dd_stat = 0;
}

/*
 * telldir
 *
 * Returns the "position" in the "directory stream" which can be used with
 * seekdir to go back to an old entry. We simply return the value in stat.
 */
long telldir (DIR * dirp)
{
  errno = 0;

  if (!dirp)
    {
      errno = EFAULT;
      return -1;
    }
  return dirp->dd_stat;
}

/*
 * seekdir
 *
 * Seek to an entry previously returned by telldir. We rewind the directory
 * and call readdir repeatedly until either dd_stat is the position number
 * or -1 (off the end). This is not perfect, in that the directory may
 * have changed while we weren't looking. But that is probably the case with
 * any such system.
 */
void seekdir (DIR * dirp, long lPos)
{
  errno = 0;

  if (!dirp)
    {
      errno = EFAULT;
      return;
    }

  if (lPos < -1)
    {
      /* Seeking to an invalid position. */
      errno = EINVAL;
      return;
    }
  else if (lPos == -1)
    {
      /* Seek past end. */
      if (dirp->dd_handle != -1)
	{
	  _findclose (dirp->dd_handle);
	}
      dirp->dd_handle = -1;
      dirp->dd_stat = -1;
    }
  else
    {
      /* Rewind and read forward to the appropriate index. */
      rewinddir (dirp);

      while ((dirp->dd_stat < lPos) && readdir (dirp))
	;
    }
}

#endif

// End of replacement for libstd


long _sol_file_handle;


DIR *_ScriptolDir(const char *path)
{
	DIR *dir;
	dir = opendir(path);
	return dir;
}

// Wrapper to stdlib

DIR *open_dir(const char *path)
{
  return opendir(path);
}

text read_dir(DIR *d)
{
  struct dirent *entry;
  if((entry = readdir(d)) == NULL) return text("");
  return text(entry->d_name);
}  


void rewind_dir(DIR *d)
{
   rewinddir(d);
   return;
}

void close_dir(DIR *d)
{
  if (closedir(d) != 0)
     runtimeError("Unable to close directory");
  return;   
}

// end of wrapper

solarray scandir(const char *name)
{
  solarray a = new solarray();
  DIR *d = opendir(name);
  rewinddir(d);
  
  struct dirent *entry;
  
  while (1)
  {
    entry = readdir(d);
    if(entry == NULL) break; 
    text name = text(entry->d_name);
    a.push(name);
  }

  closedir(d);
  return a;
}



text getcwd()
{
 char buffer[256];

 getcwd(buffer, 255);
 return text(buffer);
}


bool _sol_chdir(const char *path) { return chdir(path) == 0; }
bool _sol_rmdir(const char *path) { return rmdir(path) == 0; }
bool _sol_mkdir(const char *path)
{
#ifdef UNIX
  return mkdir(path, 255) == 0;
#else
    return mkdir(path) == 0;
#endif
}

bool _sol_ren(const char *old, const char *name)
{
  return rename(old, name) == 0;   // win only
}


bool _sol_isfile(const char *name)
{
	FILE *f;
	DIR *dir = NULL;
	if(opendir(name) != NULL)
	{
		closedir(dir);
		return false;
	}
	if((f = fopen(name, "r")) != NULL)
	{
		fclose(f);
		return true;
	}
	return false;
}

bool _sol_isdir(const char *name)
{
	DIR *dir = NULL;

	if(opendir(name) != NULL)
	{
		closedir(dir);
		return true;
	}
	return false;
}

bool _sol_unlink(const char *name)
{
	return  unlink(name) == 0;
}
