/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <ctype.h>
#include <sys/stat.h>
#include <time.h>
#include <math.h>
#include <fcntl.h>
#include <dirent.h>

#include <iostream>
#include <fstream>
#include <vector>

#ifdef UNIX
#include <unistd.h>
#else
#include <io.h>
#endif

#include "text.hpp"
#include "intarr.hpp"
#include "textarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"
#include "fun.hpp"


//===== file

void outputFileError(const char *name) { fprintf(stderr, "Cannot open %s\n", name); }

FILE *fopen(text t, text m) { return _sol_open(t.data, m.data); }
FILE *fopen(text t, const char *m) { return _sol_open(t.data, m); }

FILE *_sol_open(const char *s, const char *m)
{
  FILE *x = fopen(s, m);
  dyn::error.set( (x == NULL) ? false : true );
  return x;
}

void fseek(FILE *f, int pos)
{
  fseek(f, pos, 0);
}

text fgets(FILE *f)
{
  char buffer[2048];
  char *ptr;
  ptr = fgets((char *) buffer, 2048, f);
  if(ptr == NULL) return text(false);
  return text(buffer);
}

char *fread(FILE *f, int size)
{
  char *buffer = (char *) new char[size];	
  fread(buffer, size, 1, f);
  return buffer;
}

int fwrite(FILE *f, const text &t) { return fwrite(f, t.data); }
int fwrite(FILE *f, const char *s)
{
  return fwrite(s, strlen(s), 1, f);
}

solarray _sol_ftoarr(char *name) {
    solarray res = new solarray();
    res.load(name);
    return res;
}


text _sol_ftostr(char *name) {
    std::ifstream input(name);
    text t = "";
	for(std::string line; getline(input, line); ) {
		t += line.c_str();
	}
	input.close();
    return t;    
}


//===== dir

bool file_exists(const text &name)
{
  return (access(name.data, 0) == 0);
}

bool file_exists(const char *name)
{
 return (access(name, 0) == 0);
}


text filetype(const char *name)
{
 FILE *f;
 DIR *dir = NULL;

 if(opendir(name) != NULL)
 {
   closedir(dir);
   return text("dir");
 }

 if((f = fopen(name, "r")) != NULL)
 {
   fclose(f);
   return text("file");
 }

 return text("unknow");
}

struct stat _sol_stat;
bool _sol_filestat(const char *name)
{
   FILE *stream;
   if((stream = fopen(name, "r")) == NULL)
   {
     outputFileError(name);
     return false;
   }
   fstat(fileno(stream), &_sol_stat);
   fclose(stream);
   return true;
}

double filesize(const char *name)
{
   if (!_sol_filestat(name)) return 0;
   return (double) _sol_stat.st_size;
}

int filemtime(const char *name)   // last modification time
{
   if (!_sol_filestat(name)) return 0;
   return (int) _sol_stat.st_mtime;
}


const char *date(const char *format, int t)
{
	char cd, cu;
	time_t tt = (time_t) t;
	struct tm *intvals;
	char *str = ctime(& tt);
	char *newstr = (char *) new char[26];
	
	intvals = localtime(&tt);	
	int month = intvals->tm_mon + 1;
	
	// the year

	newstr[0] = str[20];
	newstr[1] = str[21];
	newstr[2] = str[22];
	newstr[3] = str[23];
	
	newstr[4] = '-';
	
	// the month
	
	if(month> 9) 
	{	
		cd = '1'; 
		cu = '0' + (month - 10);	
	}	
	else 
	{	
		cd = '0';
		cu = '0' + month;
	}
	
	newstr[5] = cd;
	newstr[6] = cu;
	
	newstr[7] = '-';
	
	// the day
	
	newstr[8] = str[8];
	newstr[9] = str[9];
	
	newstr[10] = 'T';
	
	// hour, minutes, second
	
	newstr[11] = str[11];
	newstr[12] = str[12];
	newstr[13] = ':';
	newstr[14] = str[14];
	newstr[15] = str[15];
	newstr[16] = ':';
	newstr[17] = str[17];
	newstr[18] = str[18];
	
	newstr[19] = '+';
	newstr[20] = '0';
	newstr[21] = '0';
	newstr[22] = ':';
	newstr[23] = '0';
	newstr[24] = '0';
	newstr[25] = '\0';
	
   return newstr;
}


void require(const char *s) { return; }
void require_once(const char *s) { return; }
