/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <stdarg.h>
#include <cstring>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <limits.h>

#include "text.hpp"
#include "textarr.hpp"


/** class _arr32
   Base class for text array
 */


int _default_capacity32 = 16;

//=====  Constructors

/**
   Basic constructor
  */
  
_arr32::_arr32()           { set(); }

_arr32::_arr32(int n)      { set(n); }

_arr32::_arr32(char *a[])  { set(a); }

_arr32::_arr32(_arr32 *a)  { set(a); }

_arr32::_arr32(const _arr32 &a) { set(a); }

_arr32::_arr32(int n, const char *s) {
  set(n);
  push(s);
}

_arr32::_arr32(const char *s) {
  set();
  push(s);
}


_arr32::_arr32(int n, text &x) {
  set(n);
  push(x.data);
}


//===== Utilities

void _arr32::clear() {
  set(_default_capacity32);
}

void _arr32::resize(int si) {
  if(si >= len)  {
    if(si >= capacity)  {
       capacity = si;
       data = (char **) realloc(data, capacity * sizeof(char *));
    }
  }
}

char *_arr32::toStr()
{
   text t = toText();
   return t.data;
}

int _arr32::compare(char *a, char *b) {
    if(text::isNumber(a)) {
        if(text::isNumber(b)) {
            int i = (int) text::toNumber(a);
            int j = (int) text::toNumber(b);
            if(i < j) return -1;
            if(i > j) return 1;
            return 0;
        }    
    }
    return strcmp(a, b);
}


void _arr32::pack() {
  int i, j;

  j = 1;
  for(i = 0; i < len; i++)  {
    if(data[i] == NULL)
    {
      while(data[j] == NULL) j++;
      if(j >= len)  {
        len = i;
        return;
      }
      data[i] = data[j];
    }
    j++;
  }
  len = i;
  return;
}

void _arr32::set() { 
    set(_default_capacity32); 
}

void _arr32::set(int si) {
    data = (char **)  new char[si * sizeof(char *)];
    capacity = si;
    len = 0;
    indexer = 0;
}


void _arr32::set(char *x) {
    set();
    data[0] = x;
    len++;
}


void _arr32::set(const _arr32 &model) {
    set(model.capacity); 

    for(int i  = 0; i < model.len; i++) {
      push(model.data[i]);
    }
}

void _arr32::set(_arr32 *model) {
    set(model->capacity);
    for(int i  = 0; i < model->len; i++)  {
      push(model->data[i]);
    }
}


void _arr32::set(char *x[]) {
  int i = 0;
  set();
  while(x[i] != NULL) push(x[i++]);
}


//===== GET

char * & _arr32::operator[](int i)  {
   if(i >= len)  {
     resize(i);
     data[i] = (char *) NULL;
   }
   return (char *&) data[i];
 }

char * & _arr32::operator[](const char *s)  {
   int k = atoi(s);
   if(k >= len) {
    resize(k);
    data[k] = NULL;
   }
   return (char *&) data[ k ];
 }

char * _arr32::at(int n) {
   if(n < len) return data[n];
   return "";
 }

char *  _arr32::at(text t) { return at(t.data); }
char * _arr32::at(const char *a) {
  if(text::isNumber(a))  {
   int n = atoi(a);
   if(n < len) return data[n];
  }
  return "";
}


void _arr32::setAt(int i, char *a) {
   if(i < len) data[i] = a;
   else push(a);
 }


void _arr32::add(_arr32 x) {
  int l = x.size();
  for(int i = 0; i < l; i++) {
     push(x.data[i]);
  }
}


void _arr32::clone(_arr32 *y)  {
   y->set(*this);
 }


//===== OPERATORS

void _arr32::operator=(const _arr32 &a)  { set(a); }
void _arr32::operator=(_arr32 *a) { set(a); }
void _arr32::operator=(char *s) {
  set();
  push(s);
}


_arr32 _arr32::operator+(_arr32 x)  {
   _arr32 y = _arr32(this);
   y.add(x);
   return y;
 }

 _arr32 _arr32::operator+(char *x)  {
   _arr32 y = _arr32(this);
   y.push(x);
   return y;
 }

 _arr32 _arr32::operator+(text t)  {
   _arr32 y = _arr32(this);
   char *temp = (char *) new char[t.len + 1];
   memcpy(temp, t.data, t.len + 1);
   y.push(temp);
   return y;
 }


_arr32 _arr32::operator-(_arr32 b) {
   int i,j;
   int l = b.size();
   _arr32 y = _arr32(len);
   char *x;

   for(i = 0; i < len; i++) {
     x = data[i];
     for(j = 0; j  < l; j++)  {
       if(strcmp(x ,b.data[j]) == 0)  {
        goto skip;
       }
     }
     y.push(x);
skip:;
   }
   return y;
};

_arr32 _arr32::operator-(char *x) {
   _arr32 y = _arr32(len);
   for(int i = 0; i < len; i++) {
     if(strcmp(data[i] , x))  y.push(data[i]);
   }
   return y;
}


  _arr32 _arr32::operator&(_arr32 b)  {
    int l = b.size();
    int m = (l < size()) ? l : size();
    _arr32 y = _arr32(m);

    for(int i = 0; i < size(); i++)  {
        for(int j = 0; j  < l; j++)  {
            if(!strcmp(data[i], b.data[j]))  y.push(data[i]);
        }
    }
    return y;
 };


_arr32 _arr32::operator&(char *v) {
    _arr32 y = _arr32(16);
    if(findValue(v) > -1) y.push(v);
    return y;
}

 _arr32 _arr32::operator&(text t)  {
    _arr32 y = _arr32(16);   
    if(findValue(t.data) > -1) y.push(t);
    return y;
 }


_arr32  _arr32::operator|(_arr32 b) {
  int i, l = b.size();
  _arr32 y = _arr32(len + l);
  for(i = 0; i < len; i++) y.pushUnic(data[i]);
  for(i = 0; i < l; i++) y.pushUnic(b.data[i]);
  return y;
}

_arr32 _arr32::operator|(char * b)  {
   _arr32 y = _arr32(this);
   y.pushUnic(b);
   return y;
 }

_arr32 _arr32::operator|(text t) {
   _arr32 y = _arr32(this);
   char *temp = (char *) new char[t.len + 1];
   memcpy(temp, t.data, t.len + 1);
   y.pushUnic(temp);
   return y;
 }


_arr32 _arr32::operator^(_arr32 b) {
  int i;
  _arr32 y = _arr32(len);

  for(i = 0; i < len; i++) {
    if(b.findValue(data[i]) == -1)
      y.pushUnic(data[i]);
  }
  for(i = 0; i < b.size(); i++) {
    if(findValue(b.data[i]) == -1)
      y.pushUnic(b.data[i]);
  }
  return y;
}

void _arr32::operator+=(_arr32 x) { add(x); }
void _arr32::operator+=(char *x)   { push(x);}


 void _arr32::operator-=(_arr32 x)  {
   _arr32 y = _arr32(this);
   set(y.operator-(x));
 }

 void _arr32::operator-=(char * x)  {
   _arr32 y = _arr32(this);
   set(y.operator-(x));
 }

void _arr32::operator&=(_arr32 b) {
   _arr32 y = operator&(b);
   set(&y);
 }

void _arr32::operator&=(char *v) {
   int flag = findValue(v) > -1;
   clear();
   if(flag) push(v);
 }


void _arr32::operator|=(_arr32 b) {
   _arr32  y = operator|(b);
   set(&y);
 }

void _arr32::operator|=(char *v) {
   pushUnic(v);
}

void _arr32::operator^=(_arr32 b) {
   _arr32  y = operator^(b);
   set(&y);
 }


bool _arr32::operator==(_arr32 b) {
   if(b.size() != len) return false;
   for(int i = 0; i < len; i++) {
     if(compare(data[i] , b.data[i])) return false;
   }
   return true;
 }

 bool  _arr32::operator==(bool x)  { return (size() != 0) == x; }

 bool _arr32::operator<(_arr32 a)  {
   int l = a.size();
   if(len < l) return true;
   if(len > l) return false;
   for(int i = 0; i < l; i++) {
     int r = compare(data[i] , a.data[i]);
     if (r > 0) return false;
     if (r < 0) return true;
   }
   return false;
 }

 bool _arr32::operator>(_arr32 a) {
   int l = a.size();
   if(len < l) return false;
   if(len > l) return true;
   for(int i = 0; i < l; i++) {
     int r = compare(data[i] , a.data[i]);
     if (r < 0) return false;
     if (r > 0) return true;
   }
   return false;
 }

bool _arr32::operator!() { return len == 0; }



//===== util

 bool _arr32::in(char *d)  {
  for(int i = 0; i < size(); i++)  {
   if(!strcmp(data[i],d)) return true;
  }
  return false;
 }

/* is the textual key already in dict? */
signed int _arr32::hasStrKey(const text &s) {
  if(text::isNumber(s.data))  {
    int i = atoi(s.data);
    return len - i;
  }
  return -1;
}

signed int _arr32::findKey(const text &s) {
  if(text::isNumber(s.data))  {
    return atoi(s.data);
  }
  return -1;
}

signed int _arr32::findKey(int n) { return n; }

signed int _arr32::findValue(char *d) {
  for(int i = 0; i < size(); i++)  {
   if(!strcmp(data[i] , d)) return i;
  }
  return -1;
}

void _arr32::display() { display(0); }
void _arr32::display(int tab)
{
    char *x;
 
    std::cout << "Array of text(\n";
 
    for(int i = 0; i < len; i++)  {
        for(int j = 0; j < tab; j++) {
            std::cout << "     ";
        }
        std::cout << "   " << i << " : " << data[i] << std::endl;
    }
    for(int j = 0; j < tab; j++) {
        std::cout << "     ";
    }    
    std::cout << ")\n";
}

 //===== iterator

 void _arr32::reset() { indexer = 0; }
 bool _arr32::outOfBound() { return indexer >= size(); }
 bool _arr32::bound()      { return(++indexer < size()); }

 char * _arr32::inc()  {
    if(++indexer < size()) { return data[indexer]; }
    indexer = size();
    return NULL;
 }

 char * _arr32::dec() {
    if(--indexer >= 0) return data[indexer];
    indexer = 0;
    return NULL;
 }

 char * _arr32::current() { return value(indexer); }
 char * _arr32::value() { return value(indexer); }
 char * _arr32::value(int i)  {
    if(i >=0 && i < size())  return data[i];
    return NULL;
 }

 int _arr32::key() {
   if(indexer >= 0 && indexer < size())  return indexer;
   return -1;
  }

 int _arr32::key(int i)
 {
   if(i >= 0 && i < size())  return i;
   return -1;
 }

 int _arr32::index()
 {
   if(indexer >=0 && indexer < size()) return indexer;
   return -1;
 }

char * _arr32::begin()
 {
   indexer = 0;
   if(size() > 0) return data[ 0 ];
   return NULL;
}

char * _arr32::end()
{
   indexer = 1;
   if(size() > 0)
   {
     indexer = size() -1;
     return data[indexer];
   }
   return NULL;
}

//===== stack


 void _arr32::push(char *a) {
    if(len >= capacity) {
        resize(capacity + 16);    
    }
    int l = strlen(a) + 1;
    data[len] = new char[l];
    memcpy(data[len], a, l);
    len++;
 }


 void _arr32::pushUnic(char *v)  {
     if(findValue(v) == -1)  {
       push(v);
     }
 }

 char * _arr32::pop()  {
   indexer = 0;
   if(!empty())  {
     len --;
     return data[len];
   }
   return NULL;
 }

char * _arr32::shift() {
  if(empty())  return NULL;
  else  {
   char *x = data[0];
   erase(0);
   return x;
  }
}

void _arr32::unshift(char *a) {
  insert(0, a);
}

void _arr32::insert(int n, char *d) {
  if(n < 0) n += len;
  if(n < 0) return;
  if(n >= len) {
    push(d);
    return;
  }

  char *x = data[len - 1];
  for(int i = len - 1; i > n; i--) {
    data[i] = data[i - 1];
  }
  data[n] = d;
  push(x);
}

void _arr32::insert(int n, _arr32 d) {
  if(n < 0) n += len;
  if(n < 0) return;
  if(n >= len) {
    add(d);
    return;
  }

  int si = len + d.size();

  _arr32 x = _arr32(si);
  if(n > 0) x.add(slice(0, n, 0));
  x.add(d);
  x.add(slice(n));
  set(x);
}

//===== RANGE


 // erase from first to last included
 void _arr32::erase(int first, int last)  {
   int k, l;

   if(first > len) return;
   if(last >= len)  {
      len = first + 1;
      return;
   }

   l = len;
   
   int nl = len - (last - first + 1);
   if(nl < 0) nl = 0;
   k = last;

   while(first < len) {
     k++;
     if(k < l) {
       data[first] = data[k];
     }
     else {
       data[first] = NULL;
     }
     first++;
   }
   len = nl;
 }

 void _arr32::erase(int pos)  {
    if(pos >= len) return;
    len--;
    while(pos < len) {
       data[pos] = data[pos + 1];
       pos++;
    }
    data[len] = NULL;
    return;
 }

 _arr32 _arr32::slice(int f, int l, int incl) {
   _arr32 x;

   l += incl;
   l--;
   if(l < 0) l += len;
   if(l < 0) goto bye;
   if(l >= len) l = len - 1;

   if(f < 0) f += len;
   if(f < 0) f = 0;
   if(l < f) goto bye;
   if(f >= len) goto bye;

   x = _arr32(l - f + 1);

   if(empty()) goto bye;

   while(f <= l) {
     x.push(data[f++]);
   }
bye:
   return x;
 }

 void _arr32::splice(int b, int e) {
   if(b < 0) b += len;
   if(b < 0) b = 0;
   if(e < 0) e += len;
   if(e < 0) return;
   if(e >= len) e = len - 1;
   if(e < b) return;
   erase(b, e);
 }

 void _arr32::splice(int b, int e, int inc) {
   splice(b, e + inc - 1);
 }

 void _arr32::splice(int f)  { erase(f); }

 void _arr32::splice(int i, int last, _arr32 a, int incl)  {
  if(empty())   { operator=(a); return; }
  if(a.empty()) { erase(i, last + incl - 1); return; }
  if(last < 0) last += len;
  if(i < 0) i += len;
  if(i < 0) i = 0;
  last += incl;
  last--;
  if(last >= len) last = len - 1;
  if(i >= len) {
    add(a);
    return;
  }
  if(last < i) return;
  if(last < 0) return;
  if((last + 1) == i)  {
    insert(i, a);       // providing not a.empty() as tested above
    return;
  }
  if(a.empty())  {
    erase(i, last);
    return;
  }

  int j = 0;
  while((i <= last) && (j < a.size())) {
    data[i] = a.data[j];
    i++;
    j++;
  }
  if(last >= i) erase(i, last);
  if(j < a.size())  {
    insert(i, a.slice(j));
  }
 }


 void _arr32::splice(int i, int last, char * a, int incl)  {
  if(i < 0) i += len;
  if(i < 0) i = 0;
  if(last < 0) last += len;
  last += incl;
  last --;
  if(last >= len) last = len - 1;
  if(last < i) return;
  if(last < 0) return;

  if(i >= len)  { 
      push(a); 
      return;
  }

  data[i] = a;
  if(last > i)  {
    erase(i + 1, last);
  }
 }

//===== TRANSFORM

/*
  if data is a not a number, search if exists further
  if yes, wait for the next occurence to push
  if not a number push it providing not already pushed
*/

_arr32 _arr32::unique() {
  _arr32 y = _arr32(len);

  for(int i = 0; i < len; i++) {
    y.pushUnic(data[i]);
  }

  return y;
}


void _arr32::swapData(int a, int b) {
  char *x;
  x = data[a];
  data[a] = data[b];
  data[b] = x;
}


void _arr32::dynSort(int first,int last) {
 int f=first;
 int l=last;
 char *item = data[(f + l) / 2];

 do  {
   while(compare(data[f] , item) < 0) f++;
   while(compare(item , data[l]) < 0) l--;
   if(f <= l)  {
     swapData(f, l);
     f++;
     l--;
   }
 } while(f <= l);

 if(first < l)  dynSort(first, l);
 if(f < last)   dynSort(f, last);
 return;
}

void _arr32::sort(int i)  { dynSort(0, len - 1); }
void _arr32::aSort(int i) { dynSort(0, len - 1); }


_arr32  _arr32::keys() {
  _arr32 x = _arr32(len);
  for(int i = 0; i < len; i++)  {
    text t = text(i);  
    x.push(t.data);
  }
  return x;
}

_arr32 _arr32::values() {
  return *this;
}

//===== UTILITIES

int _arr32::min() {
  int m = INT_MAX;
  int x = 0;

  for(int i = 0; i < len; i++) {
    x = atoi(data[i]);
    if(x < m) m = x;
  }
  return m;
}

int _arr32::max() {
    int m = INT_MIN;
    int x = 0;

    for(int i = 0; i < len; i++)  {
        x = atoi(data[i]);
        if(x > m) m = x;
    }
    return m;
}


int _arr32::sum() {
    int m = 0;
    for(int i = 0; i < len; i++)
       m += atoi(data[i]);
    return m;
}


char *_arr32::toString(const char *s) {
	text *t = new text();
    t->cat("(");
	for(int i = 0; i < len; i++) 	{
		if(i > 0) t->cat(s);
		t->cat(data[i]);
	}
	t->cat(")");
	return t->data;
}

text _arr32::join(const char *sep) {
  return text(toString(sep));
}

_arr32 _arr32::reverse() {
  _arr32 y = _arr32(len);
  for(int i = len - 1; i >= 0; i--)
   y.push(data[i]);
  return y;
}

text _arr32::toText() {
    text x = "(";
    for(int i = 0; i < len; i++)  {
        if(i > 0) x +=  ", ";
        x += text(i);
        x += ":";
        x += text(data[i]);
    }
    x += ")";
    return x;
}


std::ostream& operator<< (std::ostream& out, _arr32 &d) {
  return out << d.toStr();
}

void _arr32::load(const char *name) {
    std::ifstream input(name);
    this->clear();
	for(std::string line; getline(input, line); ) {
		push(line.c_str());
	}
	input.close(); 
	return;
}

bool _arr32::store(const char * name) {
    std::ofstream f;
    f.open(name);
	for(int i = 0; i < size(); i++)	{
	    std::stringstream out;
        out << data[i];
        f <<  out.str() << "\n";
	}
	f.close();
    return true;
}


void _arr32::boundError(const text &t) {
  std::cout << "Out of boundary " << t << std::endl;
}

